require "hash_wrap/version"

class HashWrap
  def initialize( hash )
    @hash = case hash
    when Hash
      hash.transform_keys(&:to_sym).freeze
    when HashWrap
      hash
    else
      raise "dunno how to wrap #{hash.inspect}"
    end
  end

  def method_missing(meth, *args, &blk)
    # handle both string and symbol keys
    value = @hash.fetch meth.to_s do |_key|
      @hash.fetch meth do |_key|
        super
      end
    end
    __wrap value
  end

  private def __wrap value
    case value
    when Hash
      self.class.new value
    when Array
      value.map do |item|
        __wrap item
      end
    else
      value
    end
  end

  private def __hash__; @hash end

  def == rhs
    @hash == __hash__
  end

  def each &blk
    return enum_for :each unless block_given?
    @hash.each do |k,v|
      yield [k,__wrap(v)]
    end
  end

  def to_h
    # no, you can't have a copy of this hash to mess with
    @hash.dup
  end

  def dig *parts
    @hash.dig *parts
  end

  def deconstruct
    raise NotImplementedError
  end

  def deconstruct_keys keys
    if keys.nil? || keys.empty?
      @hash
    else
      @hash.slice(*keys).map{|k,v| [k, __wrap(v)]}.to_h
    end
  end

  def to_s
    @hash.to_s
  end

  def inspect
    @hash.inspect
  end

  # have to handle this otherwise pry gets upset
  def pretty_print(pp)
    pp.text inspect
  end

  # for awesome_print
  def ai; inspect end
end

module Kernel
  private def HashWrap arg
    case arg
    when Hash; HashWrap.new arg
    else;      HashWrap.new arg.to_h
    end
  end
end
